#include <windows.h>

#include "menukeys.h"
#define MAXSTR  (255)
#define NEWARRAY(s,n)   ((s*)malloc(n*sizeof(s)))

/* FindAmp() - find the first non-literal ampersand in a string */
char*   FindAmp(char* Str)
    {
    char    *Result = 0;    /* assume failure   */

    while(Str != NULL)
        {
        Str     = strchr(Str, '&');
        if(Str != NULL)     /* if we found an ampersand     */
            {
            if(Str[1] == '&')   /* if literal '&'           */
                Str     += 2;   /* then skip it             */
            else
                {
                Result  = Str;
                break;
                }
            }
        }
    return Result;
    }

void AssignNewKey(int* KeyTakenBy, char* Text, HMENU hMenu,
        int MenuItem)
    {
    int     Pos;

    /* Find first available char. Stop at TAB if found      */
    for(Pos=0; Text[Pos] && Text[Pos] != '\t'; ++Pos)
        {
        char    Char    = Text[Pos];
        char    Lower   = (char)tolower(Char);

        /* If legal char and not already in use     */
        if(isalnum(Char) && !KeyTakenBy[Lower])
            {
            /* then assign this key         */
            KeyTakenBy[Lower]   = MenuItem + 1;

            /* make room for ampersand  */
            memmove(Text + Pos + 1, Text + Pos, 
                        lstrlen(Text) - Pos + 1);
            Text[Pos] = '&';
            break;
            }
        }
    /* This may have no effect if the string had no ampersand and
     * no good character was found, but placing it outside the for
     * loop ensures that ampersands are removed when one is 
     * initially present but has a conflict and no alternate is
     * available
     */
    ModifyMenu(hMenu, MenuItem, MF_BYPOSITION | MF_STRING, 
        GetMenuItemID(hMenu,MenuItem), Text);
}

void CleanMenuKeys(HMENU hMenu)
{
    int*    KeyTakenBy;
    char*   ItemText;
    char*   PtrAmpersand;
    char    MnemonicChar;
    int     MenuItem;
    int     ItemCount;

    /* First, allocate memory dynamically to avoid using lots of
     * stack during recursion.
     */
    if(hMenu == NULL)
        return;
    else if((KeyTakenBy=NEWARRAY(int,256)) == 0)
        return;
    else if((ItemText=NEWARRAY(char,MAXSTR+1)) == 0)
        {
        free(KeyTakenBy);
        return;
        }
    memset(KeyTakenBy, 0, sizeof(int)*256);

    ItemCount   = GetMenuItemCount(hMenu);

    //Search menu backwards so that 'top' items get to keep keys
    for(MenuItem = ItemCount - 1; MenuItem >= 0; --MenuItem)
        {
        GetMenuString(hMenu, MenuItem, ItemText,
                MAXSTR, MF_BYPOSITION);

        /* Find first '&'   */
        PtrAmpersand = FindAmp(ItemText);
        if(PtrAmpersand)
            {
            MnemonicChar = PtrAmpersand[1];
            KeyTakenBy[tolower(MnemonicChar)] = MenuItem + 1;
            }
        }
    /* Now assign for all with no key and dupes */
    for(MenuItem = 0; MenuItem < ItemCount; ++MenuItem)
        {
        GetMenuString(hMenu, MenuItem, ItemText, 
            MAXSTR - 1, MF_BYPOSITION);
        /* Separators return empty strings      */
        if(ItemText[0] == '\0')
            continue;

        /* Find first '&'           */
        PtrAmpersand = FindAmp(ItemText);

        /* if '&' is last char in string    */
        if(PtrAmpersand && PtrAmpersand[1] == '\0')
            {
            /* then remove it       */
            *PtrAmpersand   = '\0';
            PtrAmpersand    = NULL;
            }

        if(PtrAmpersand == NULL)
            AssignNewKey(KeyTakenBy, ItemText, hMenu, MenuItem);
        else
            {
            /* Check for duplicates     */
            MnemonicChar = PtrAmpersand[1];
            if(KeyTakenBy[tolower(MnemonicChar)] != (MenuItem + 1))
                {
                /* this is a dupe, and must be cleaned up   */
                /* First, remove ampersand                  */
                memmove(PtrAmpersand, PtrAmpersand + 1, 
                            lstrlen(PtrAmpersand));
                /* Now find new one, just like it never had one */
                /* whether successful or not, change it         */
                AssignNewKey(KeyTakenBy, ItemText, hMenu, MenuItem);
                }
            }
        }

    /* free up locally allocated memory     */
    free(KeyTakenBy);
    free(ItemText);

    //Now recurse for all submenus
    for(MenuItem = 0; MenuItem < ItemCount; ++MenuItem)
        CleanMenuKeys( GetSubMenu(hMenu, MenuItem) );
    }

/* Alternate input interface        */
void CleanMenuKeysWindow(HWND hWnd)
    {
    CleanMenuKeys(GetMenu(hWnd));
    }
